// -----------------------------------------------------------------
// deckview - display and edit decks in 1130 simulator binary format
// -----------------------------------------------------------------

#include <windows.h>
#include <windowsx.h>
#include <commdlg.h>
#include <dlgs.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "deckview.h"

#define ID_ENCODING_CHOICE	10100

#define DEFAULT_ENCODING_FILE	"cardcode.tbl"
#define DEFAULT_ENCODING_NAME	"029"

// -----------------------------------------------------------------
// NOTICES:

#define COPYRIGHT "Copyright 2003 Quarterbyte Systems, Inc."

// Licensed under the GNU General Public License, see http://www.gnu.org/copyleft/gpl.txt.
//
// Copying and distribution of this software is permitted in any medium, provided
// these notices are preserved.
// -----------------------------------------------------------------

#define REGISTRY_KEY "Software\\Quarterbyte\\DeckView"

#define BETWEEN(v,a,b) ((v) >= (a) && (v) <= (b))

#undef assert
#ifdef NDEBUG
    #define assert(exp)	((void) 0)
#else
	void _my_assert(char *expr, char *filename, unsigned lineno);
    #define assert(exp) (void)( (exp) || (_my_assert(#exp, __FILE__, __LINE__), 0) )
#endif

// -----------------------------------------------------------------
// Default encoding tables
// -----------------------------------------------------------------

#include "cr_encoding.h"

// -----------------------------------------------------------------
// Global Variables
// -----------------------------------------------------------------

#define MAX_FINDWHAT 81

#define MAXCARDS  2000

typedef struct tag_deckinfo {
	HWND hDlg;
	USHORT cardbuf[MAXCARDS*80];
	int ncards;
	int cardno;
	BOOL changed;
	BOOL named;
	BOOL editable;
	char curname[MAX_PATH];						// default filename
	HWND hwDeck;
	HWND hwCard;								// window handle of rectangle control that we use for card display
	HMENU hPopup;								// encoding popup menu
	HANDLE hClipboardData;						// data memorized for clipboard if any
	int	   nClipboardLines;						// number of lines stored on clipboard
	ENCODING *cur_encoding;						// current encoding table
	BOOL found, down;							// TRUE if previous Find operation was successful, and if so, direction
	char lpstrFindWhat[MAX_FINDWHAT];			// previous search string
} DECKINFO;

HINSTANCE ghInstance;							// instance handle of program
HBRUSH bManilla = NULL;							// handle to a brush the color of a manilla punched card
HBRUSH bBlack   = NULL;							// handle to a red pen
HACCEL hAccel   = NULL;							// keyboard accelerators Ctrl+X etc
WNDPROC oldRectProc;							// original window procedure for the IDC_CARD display control
WNDPROC oldListProc;							// original window procedure for the IDC_DECK display list
UINT cfCardFormat = 0;
int nwindows = 0;
char encoding_name[50];							// default encoding, from registry
char encoding_file[MAX_PATH];
char encoding_file_path[MAX_PATH];
HWND hwTop  = NULL;								// most recently selected window
HWND hwFind = NULL;
UINT msgFindText;

// -----------------------------------------------------------------
// Prototypes
// -----------------------------------------------------------------

BOOL CALLBACK	DeckViewDlgProc				(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL			DeckView_OnInitDialog		(HWND hDlg, HWND hwndFocus, LPARAM lParam);
void			DeckView_OnClose			(HWND hDlg);
void			DeckView_OnDestroy			(HWND hDlg);
void 			DeckView_OnCommand			(HWND hDlg, int id, HWND hWndCtl, UINT codeNotify);
void			DeckView_OnInitMenuPopup	(HWND hDlg, HMENU hMenu, UINT item, BOOL fSystemMenu);
int				DeckView_OnRenderFormat		(HWND hDlg, UINT cfFormat);
void			DeckView_OnRenderAllFormats	(HWND hDlg);
void			DeckView_OnDestroyClipboard	(HWND hDlg);
BOOL			DeckView_OnFindText			(HWND hDlg, WPARAM wParam, FINDREPLACE *ft);

BOOL			DeckView_Cmd_File_Save		(HWND hDlg, DECKINFO *di);
BOOL			DeckView_Cmd_File_SaveAs	(HWND hDlg, DECKINFO *di);
BOOL			DeckView_Cmd_File_New		(HWND hDlg, DECKINFO *di);
BOOL			DeckView_Cmd_File_Open		(HWND hDlg, DECKINFO *di);
BOOL			DeckView_Cmd_Edit_Cut		(HWND hDlg, DECKINFO *di);
BOOL			DeckView_Cmd_Edit_Copy		(HWND hDlg, DECKINFO *di);
BOOL			DeckView_Cmd_Edit_Paste		(HWND hDlg, DECKINFO *di);
BOOL			DeckView_Cmd_Edit_Delete	(HWND hDlg, DECKINFO *di);
BOOL			DeckView_Cmd_Edit_SelectAll	(HWND hDlg, DECKINFO *di);
BOOL			DeckView_Cmd_Edit_Find		(HWND hDlg, DECKINFO *di);
BOOL			DeckView_Cmd_Edit_FindNext	(HWND hDlg, DECKINFO *di);
BOOL			DeckView_Cmd_NextSeq		(HWND hDlg, DECKINFO *di);
BOOL			DeckView_Cmd_Goto			(HWND hDlg, DECKINFO *di);

BOOL			init						(char *fname);
BOOL			open_window					(char *fname, DECKINFO *old_window);
BOOL			load_deck					(HWND hDlg, DECKINFO *di, char *fname);
BOOL			write_deck					(HWND hDlg, DECKINFO *di, char *fname, BOOL as_text);
BOOL			query_close_file			(HWND hDlg, DECKINFO *di);
void			set_window_title			(DECKINFO *di);
void			CenterMe					(HWND hWnd);
void			load_reg_bool				(HKEY hKey, LPSTR valname, BOOL *val, BOOL default_val);
void			load_reg_str				(HKEY hKey, LPSTR valname, LPSTR val, int maxval, LPSTR default_val);
void			save_reg_bool				(HKEY hKey, LPSTR valname, BOOL val);
void			save_reg_str				(HKEY hKey, LPSTR valname, LPSTR val);
void			get_defaults				(char *dir, int s_dir);
void			save_defaults				(void);
char *			astring						(char *str);
char *			skipbl						(char *str);
void			paint_card					(DECKINFO *di);
void			show_card					(HWND hDlg, DECKINFO *di);
void			convert_card				(ENCODING *encoding, USHORT *card, char *cardtext);
BOOL			copy_selection_to_clipboard (HWND hDlg, DECKINFO *di);
BOOL			copy_clipboard_to_deck		(HWND hDlg, DECKINFO *di);
BOOL			delete_selection			(HWND hDlg, DECKINFO *di);
LRESULT CALLBACK MyRectProc					(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK MyListProc					(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void			bail						(char *msg);
int				get_selected_items			(DECKINFO *di, int *selindex, int maxsel);
void			modifying_deck				(DECKINFO *di, BOOL modifying);
int				render_text					(DECKINFO *di);
void			redisplay_deck				(DECKINFO *di);
void			add_end_of_deck				(DECKINFO *di);
void			select_encoding				(DECKINFO *di, int i);
BOOL			jump_to_line				(DECKINFO *di, int i);

// -----------------------------------------------------------------
// WinMain - main program
// -----------------------------------------------------------------

int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int nCmdShow)
{
	MSG msg;

	ghInstance = hInstance;

	if (! init(lpszCmdLine))
		return 1;

	while (GetMessage(&msg, NULL, 0, 0)) {					// set up message processing
		if (hAccel && hwTop)								// accelerator message? (e.g. Ctrl+C)
			if (TranslateAccelerator(hwTop, hAccel, &msg))
				continue;

		if (hwTop)
			if (IsDialogMessage(hwTop, &msg))
				continue;

		if (hwFind)
			if (IsDialogMessage(hwFind, &msg))
				continue;

		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	
	return 0;
}

BOOL init (char *fname)
{
	char *c, hold, dir[MAX_PATH];

	msgFindText = RegisterWindowMessage("commdlg_FindReplace");

	hAccel   = LoadAccelerators(ghInstance, MAKEINTRESOURCE(IDR_ACCEL_DECKVIEW));

	bManilla = CreateSolidBrush(RGB(234,241,163)); 			// create brush for painting card image
	bBlack   = GetStockObject(BLACK_BRUSH);					// use stock black brush

	get_defaults(dir, sizeof(dir));							// get previous settings from registry

	if ((cfCardFormat = RegisterClipboardFormat("DECKVIEW_BINARY_DATA")) == 0) {
		bail("Unable to register clipboard format");
		return FALSE;
	}

	if (fname != NULL) {
		while (*fname && *fname <= ' ')						// if filename specified, ignore if blank
			fname++;										// (e.g. from command line)

		if (*fname == '"') {								// if encased in quotes, remote quotes
			fname++;
			for (c = fname; *c && *c != '"'; c++)
				;
			*c = '\0';

		}

		if (! *fname)										// if string was blank, pretend we never saw it
			fname = NULL;
	}
															// if a filename is specified on the command line,
	if (fname != NULL) {
		if (fname[0] == '\\' || fname[1] == ':') {			// if we have an absolute path, change to it
			if ((c = strrchr(fname, '\\')) == NULL)			// terminate at the backslash
				if ((c = strchr(fname, ':')) != NULL)				
					c++;									// or if it contains just a drive, terminate after the colon

			if (c != NULL) {								// truncate string and change to specified directory
				hold = *c;
				*c = '\0';
				SetCurrentDirectory(fname);
				*c = hold;
			}													
		}													// otherwise, don't, as filename is relative to current dir
	}
	else if (*dir)											// if no filename specified on command line
		SetCurrentDirectory(dir);							// restore last-used directory
			
	read_encoding_tables(encoding_file_path, GetDesktopWindow());

	return open_window(fname, NULL);
}

BOOL open_window (char *fname, DECKINFO *old_window)
{
	DECKINFO *di;
	HICON hIcon;
	HMENU hMenu;
	int i;

	if ((di = (DECKINFO *) malloc(sizeof(DECKINFO))) == NULL) {
		MessageBox(IsWindow(hwTop) ? hwTop : GetDesktopWindow(), "Insufficient memory", NULL, MB_OK);
		return FALSE;
	}

	ZeroMemory(di, sizeof(*di));

	if (CreateDialogParam(ghInstance, MAKEINTRESOURCE(IDR_DLG_DECKVIEW), NULL, DeckViewDlgProc, (LPARAM) di) == NULL) {
		MessageBox(IsWindow(hwTop) ? hwTop : GetDesktopWindow(), "Unable to create window", NULL, MB_OK);
		free(di);
		return FALSE;
	}

	if (nwindows == 0) {
		if ((hIcon = LoadIcon(ghInstance, MAKEINTRESOURCE(IDR_ICON_DECKVIEW))) != NULL)
			SetClassLong(di->hDlg, GCL_HICON, (LONG) hIcon);		// set the Dialog class icon to our program's icon
	}

	nwindows++;

	select_encoding(di, 0);											// pick default as something reasonable: first encoding

	di->hPopup = CreatePopupMenu();
	for (i = 0; i < nencodings; i++) {
		AppendMenu(di->hPopup, MF_STRING, ID_ENCODING_CHOICE+i, encoding[i].name);
		if (old_window == NULL && strcmpi(encoding_name, encoding[i].name) == 0)
			select_encoding(di, i);
		else if (old_window != NULL && old_window->cur_encoding == &encoding[i])
			select_encoding(di, i);
	}

	hMenu = GetMenu(di->hDlg);
	hMenu = GetSubMenu(hMenu, 2); 
	AppendMenu(hMenu, MF_POPUP, (UINT) di->hPopup, "En&coding");

	if (fname != NULL)
		load_deck(di->hDlg, di, fname);

	add_end_of_deck(di);
	return TRUE;
}

// -----------------------------------------------------------------
// DeckViewDlgProc - dialog procedure for the main DeckView window
// -----------------------------------------------------------------

BOOL CALLBACK DeckViewDlgProc (HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {
		case WM_INITDIALOG:
			return HANDLE_WM_INITDIALOG(hDlg, wParam, lParam, DeckView_OnInitDialog);

		case WM_COMMAND:
			HANDLE_WM_COMMAND(hDlg, wParam, lParam, DeckView_OnCommand);
			break;

		case WM_CLOSE:
			HANDLE_WM_CLOSE(hDlg, wParam, lParam, DeckView_OnClose);
			break;

		case WM_DESTROY:
			HANDLE_WM_DESTROY(hDlg, wParam, lParam, DeckView_OnDestroy);
			return FALSE;						// let default dialog procedure perform its own cleanup

		case WM_INITMENUPOPUP:
			HANDLE_WM_INITMENUPOPUP(hDlg, wParam, lParam, DeckView_OnInitMenuPopup);
			break;

		case WM_ACTIVATE:
			hwTop = hDlg;
			return FALSE;						// pretend we did not handle this message

		case WM_RENDERFORMAT:
			return HANDLE_WM_RENDERFORMAT(hDlg, wParam, lParam, DeckView_OnRenderFormat);

		case WM_RENDERALLFORMATS:
			HANDLE_WM_RENDERALLFORMATS(hDlg, wParam, lParam, DeckView_OnRenderAllFormats);
			break;

		case WM_DESTROYCLIPBOARD:
			HANDLE_WM_DESTROYCLIPBOARD(hDlg, wParam, lParam, DeckView_OnDestroyClipboard);
			break;

		default:
			if (uMsg == msgFindText)			// notification from FindText modeless dialog
				return DeckView_OnFindText(hDlg, wParam, (FINDREPLACE *) lParam);

			return FALSE;
	}

	return TRUE;
}

// -----------------------------------------------------------------
// DeckView_OnInitDialog - initialize program
// -----------------------------------------------------------------

BOOL DeckView_OnInitDialog (HWND hDlg, HWND hwndFocus, LPARAM lParam)
{
	DECKINFO *di;

	di = (DECKINFO *) lParam;
	SetWindowLong(hDlg, DWL_USER, (LPARAM) di);
	di->hDlg = hDlg;

	CenterMe(hDlg);											// center dialog on screen

	di->hwCard = GetDlgItem(hDlg, IDC_CARD);				// get handle to rectangle that will contain card picture
	di->hwDeck = GetDlgItem(hDlg, IDC_DECK);
	oldRectProc = (WNDPROC) GetWindowLong(di->hwCard, GWL_WNDPROC);	// superclass this control so we can intercept WM_PAINT messages
	SetWindowLong(di->hwCard, GWL_WNDPROC, (LONG) (WNDPROC) MyRectProc);
															// set controls to match initial settings
	SetDlgItemText(hDlg, IDC_CARDNO, "");
	SetDlgItemText(hDlg, IDC_OF,     "");
	SetDlgItemText(hDlg, IDC_NCARDS, "");
	di->changed = di->named = FALSE;
	di->cardno  = 0;
	strcpy(di->curname, "unnamed.dat");
	di->editable = TRUE;

	oldListProc = (WNDPROC) GetWindowLong(di->hwDeck, GWL_WNDPROC);	// superclass the list so we can intercept arrow characters
	SetWindowLong(di->hwDeck, GWL_WNDPROC, (LONG) (WNDPROC) MyListProc);

	return TRUE;
}

// handle_arrow_key - process arrow key in list when shift is up. Cancels any selection
// and moves up and down (from current line, or top or bottom of selection if there was one)

void handle_arrow_key (HWND hWnd, int dir, BOOL shift)
{
	HWND hDlg = GetParent(hWnd);
	DECKINFO *di = (DECKINFO *) GetWindowLong(hDlg, DWL_USER);
	int oldindex, newindex, selindex[MAXCARDS], nsel;
	BOOL newsel;

	oldindex = SendMessage(hWnd, LB_GETCARETINDEX, 0, 0);
	newindex = oldindex + dir;
	newindex = max(0, min(newindex, di->ncards));

	if (shift) {
		if (newindex == oldindex)													// prevents funny business at top, BUT
			return;																	// you can't select the top line by
																					// moving up from below. Guess that's OK.
		if (SendMessage(hWnd, LB_GETSEL, newindex, 0)) {
			SendMessage(hWnd, LB_SETSEL, FALSE, newindex);
		}
		else {
			newsel = ! SendMessage(hWnd, LB_GETSEL, oldindex, 0);
			if (BETWEEN(oldindex, 0, di->ncards-1) || ! newsel)
				SendMessage(hWnd, LB_SETSEL, newsel, oldindex);
		}
	}
	else {																			// shift off cancels selection
		if (SendMessage(hWnd, LB_GETSELCOUNT, 0, 0) > 0) {
			nsel = get_selected_items(di, selindex, MAXCARDS);
			newindex = (dir < 0) ? (selindex[0]-1) : (selindex[nsel-1]+1);
			SendMessage(hWnd, LB_SELITEMRANGEEX, selindex[nsel-1], selindex[0]);	// unselect all
		}
	}

	SendMessage(hWnd, LB_SETCARETINDEX, newindex, FALSE);							// move to new line
	di->cardno = newindex;															// force update since
	show_card(hDlg, di);															// no LBN_SELCHANGE will be sent
}

LRESULT CALLBACK MyListProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {
		case WM_KEYDOWN:
		case WM_KEYUP:
			if (wParam == VK_UP || wParam == VK_DOWN) {	// make arrow keys work like they do in edit boxes
				if (uMsg == WM_KEYDOWN)
					handle_arrow_key(hWnd, (wParam == VK_UP) ? -1 : +1, GetKeyState(VK_SHIFT) & 0x8000);

				return 0;
			}
			break;

		case WM_CHAR:
			if (wParam == ' ')							// don't let spacebar cause selection
				return 0;

			break;

		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
			if (GetKeyState(VK_CONTROL) & 0x8000)		// prevent ctrl+click (noncontiguous selection)
				return 0;

			break;
	}
														// all other messages processed by standard listbox handler
	return (*oldListProc)(hWnd, uMsg, wParam, lParam);
}

// -----------------------------------------------------------------
// DeckView_OnCommand - handle command messages (GUI clicks)
//
// (Note: it's not usual Windows programming style, but in this
//  program, ALL dialog values are extracted from the dialog to global
//  variables IMMEDIATELY upon their being changed by the user.
//  This made the port from the command line version easier.)
// -----------------------------------------------------------------

void DeckView_OnCommand (HWND hDlg, int id, HWND hWndCtl, UINT codeNotify)
{
	DECKINFO *di = (DECKINFO *) GetWindowLong(hDlg, DWL_USER);

	switch (id) {
		case ID_FILE_NEW:
			DeckView_Cmd_File_New(hDlg, di);
			break;
		case ID_FILE_OPEN:
			DeckView_Cmd_File_Open(hDlg, di);
			break;
		case ID_FILE_SAVE:
			DeckView_Cmd_File_Save(hDlg, di);
			break;
		case ID_FILE_SAVEAS:
			DeckView_Cmd_File_SaveAs(hDlg, di);
			break;
		case ID_FILE_EXIT:
			DeckView_OnClose(hDlg);
			break;
		case ID_EDIT_CUT:
			DeckView_Cmd_Edit_Cut(hDlg, di);
			break;
		case ID_EDIT_COPY:
			DeckView_Cmd_Edit_Copy(hDlg, di);
			break;
		case ID_EDIT_PASTE:
			DeckView_Cmd_Edit_Paste(hDlg, di);
			break;
		case ID_EDIT_DELETE:
			DeckView_Cmd_Edit_Delete(hDlg, di);
			break;
		case ID_EDIT_SELECTALL:
			DeckView_Cmd_Edit_SelectAll(hDlg, di);
			break;
		case ID_EDIT_FIND:
			DeckView_Cmd_Edit_Find(hDlg, di);
			break;
		case ID_EDIT_FINDNEXT:
			DeckView_Cmd_Edit_FindNext(hDlg, di);
			break;
		case ID_NEXTSEQ:
			DeckView_Cmd_NextSeq(hDlg, di);
			break;
		case ID_GOTO:
			DeckView_Cmd_Goto(hDlg, di);
			break;
		case IDC_DECK:
			if (codeNotify == LBN_SELCHANGE) {
				di->cardno = SendMessage(di->hwDeck, LB_GETCARETINDEX, 0, 0);
				show_card(hDlg, di);
			}
			break;
		default:
			if (BETWEEN(id, ID_ENCODING_CHOICE, ID_ENCODING_CHOICE+nencodings-1)) {
				select_encoding(di, id-ID_ENCODING_CHOICE);
				redisplay_deck(di);

				if (di->cur_encoding != NULL)						// save last-chosen name so when we exit, this is
					strcpy(encoding_name, di->cur_encoding->name);		// set as default for next time program is run
			}
			break;
	}
}

// -----------------------------------------------------------------
// DeckView_OnInitMenuPopup - update menus just prior to activation
// -----------------------------------------------------------------

void enable_menu_item (HMENU hMenu, DECKINFO *di, UINT item, BOOL enable)
{
	EnableMenuItem(hMenu, item, MF_BYCOMMAND | (enable ? MF_ENABLED : MF_GRAYED));
}

void DeckView_OnInitMenuPopup (HWND hDlg, HMENU hMenu, UINT item, BOOL fSystemMenu)
{
	DECKINFO *di = (DECKINFO *) GetWindowLong(hDlg, DWL_USER);
	BOOL is_sel;
	BOOL on_clip;
	UINT cfFormat;
	int selindex[2], n_sel;

	n_sel  = get_selected_items(di, selindex, 2);
	is_sel = n_sel > 0;

	on_clip = FALSE;
	if (OpenClipboard(hDlg)) {
		cfFormat = 0;
		while ((cfFormat = EnumClipboardFormats(cfFormat)) != 0) {
			if (cfFormat == cfCardFormat || cfFormat == CF_TEXT || cfFormat == CF_OEMTEXT) {
				on_clip = TRUE;
				break;
			}
		}
		CloseClipboard();
	}

	enable_menu_item(hMenu, di, ID_FILE_SAVE,    di->changed);
	enable_menu_item(hMenu, di, ID_FILE_SAVEAS,  TRUE);
	enable_menu_item(hMenu, di, ID_EDIT_CUT,     is_sel  && di->editable);
	enable_menu_item(hMenu, di, ID_EDIT_COPY,    is_sel);
	enable_menu_item(hMenu, di, ID_EDIT_PASTE,   on_clip && di->editable);
	enable_menu_item(hMenu, di, ID_EDIT_DELETE,  is_sel  && di->editable);
	enable_menu_item(hMenu, di, ID_NEXTSEQ,      di->ncards > 2 && n_sel < 2);		// ncards > 2 because one is the end marker
	enable_menu_item(hMenu, di, ID_GOTO, 		 di->ncards > 2);
	enable_menu_item(hMenu, di, ID_EDIT_FIND,    di->ncards > 2);
	enable_menu_item(hMenu, di, ID_EDIT_FINDNEXT,di->found);
}

// -----------------------------------------------------------------
// query_close_file - see if it's OK to close the current file
// -----------------------------------------------------------------

BOOL query_close_file (HWND hDlg, DECKINFO *di)
{
	char msg[MAX_PATH+100];

	if (! di->changed)
		return TRUE;

	sprintf(msg, "Do you want to save changes to %s?", di->curname);

	switch (MessageBox(hDlg, msg, "", MB_YESNOCANCEL|MB_ICONQUESTION)) {
		case IDYES:
			if (! DeckView_Cmd_File_Save(hDlg, di))
				return FALSE;
			break;

		case IDNO:
			break;

		case IDCANCEL:
			return FALSE;
	}

	return TRUE;
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

void set_window_title (DECKINFO *di)
{
	char title[MAX_PATH+20], *c;

	if (di->named) {
		if ((c = strrchr(di->curname, '\\')) != NULL)
			c++;
		else if ((c = strrchr(di->curname, ':')) != NULL)
			c++;
		else
			c = di->curname;

		strcpy(title, c);
		strcat(title, " - DeckView");
		SetWindowText(di->hDlg, title);
	}
	else
		SetWindowText(di->hDlg, "DeckView");
}

// -----------------------------------------------------------------
// DeckView_OnClose - shut down the current window
// -----------------------------------------------------------------

void DeckView_OnClose (HWND hDlg)
{
	if (! query_close_file(hDlg, (DECKINFO *) GetWindowLong(hDlg, DWL_USER)))
		return;

	DestroyWindow(hDlg);
}

// -----------------------------------------------------------------
// DeckView_OnDestroy - release data, perhaps shut down the program
// -----------------------------------------------------------------

void DeckView_OnDestroy (HWND hDlg)
{
	DECKINFO *di;

	if ((di = (DECKINFO *) GetWindowLong(hDlg, DWL_USER)) != NULL)
		free(di);

	if (--nwindows <= 0) {
		save_defaults();
		PostQuitMessage(0);
	}
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL DeckView_Cmd_File_Save (HWND hDlg, DECKINFO *di)
{
	if (! di->named)
		return DeckView_Cmd_File_SaveAs(hDlg, di);

	if (! di->editable)
		return FALSE;

	if (! write_deck(hDlg, di, di->curname, FALSE))
		return FALSE;

	di->changed = FALSE;
	return TRUE;
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL write_deck (HWND hDlg, DECKINFO *di, char *fname, BOOL as_text)
{
	FILE *fd;
	int i, nwritten;
	char errbuf[MAX_PATH+100], cardtext[90];
	extern int errno;
	extern char *_sys_errlist[];

	if ((fd = fopen(fname, as_text ? "w" : "wb")) == NULL) {
		sprintf(errbuf, "%s: %s", fname, _sys_errlist[errno]);
		MessageBox(hDlg, errbuf, NULL, MB_OK|MB_ICONEXCLAMATION);
		return FALSE;
	}

	if (as_text) {
		for (i = 0; i < di->ncards; i++) {
			convert_card(di->cur_encoding, &di->cardbuf[80*i], cardtext);
			if (fputs(cardtext, fd) == EOF)
				break;
			if (putc('\n', fd) == EOF)
				break;
		}
		nwritten = i;
	}
	else {
		nwritten = (int) fwrite(di->cardbuf, 80*sizeof(*di->cardbuf), di->ncards, fd);
	}

	if (nwritten != di->ncards) {
		sprintf(errbuf, "%s: %s", fname, _sys_errlist[errno]);
		fclose(fd);
		MessageBox(hDlg, errbuf, NULL, MB_OK|MB_ICONEXCLAMATION);
		return FALSE;
	}

	fclose(fd);
	return TRUE;
}

// -----------------------------------------------------------------'
// -----------------------------------------------------------------

BOOL __declspec(dllexport) CALLBACK getsave_hook_func (HWND HWnd, UINT iMsg, WPARAM wparam, LPARAM lParam)
{
	LPOFNOTIFY ofn;
	char *newext, *curext;
	char curname[MAX_PATH];

	if (iMsg == WM_NOTIFY) {
		ofn = (LPOFNOTIFY) lParam;
		if (ofn->hdr.code == CDN_TYPECHANGE) {								// if user changed file type selection
			newext = (ofn->lpOFN->nFilterIndex == 2) ? ".txt" : ".dat";		// determine correct extension
			SendMessage(ofn->hdr.hwndFrom, CDM_GETSPEC, sizeof(curname), (LPARAM) curname);
			if ((curext = strrchr(curname, '.')) != NULL) {					// if old name has an extension
				if (strcmpi(curext, newext) != 0) {							// and it's not the correct one
					strcpy(curext, newext);									// change it and update the dialog
					CommDlg_OpenSave_SetControlText(ofn->hdr.hwndFrom, edt1, curname);
				}
			}
		}
	}

	return FALSE;
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL DeckView_Cmd_File_SaveAs (HWND hDlg, DECKINFO *di)
{
	OPENFILENAME ofn;
	char namebuf[MAX_PATH], curdir[MAX_PATH];
	int i;
	BOOL as_text;

	GetCurrentDirectory(sizeof(curdir), curdir);

	strcpy(namebuf, di->curname);
	if ((i = strlen(namebuf)) > 0)
		if (namebuf[i-1] == '\\')
			namebuf[i-1] = '\0';							// dialog doesn't like it if name ends in backslash

	ZeroMemory(&ofn, sizeof(ofn));
	ofn.lStructSize       = sizeof(ofn);
	ofn.hwndOwner         = hDlg;
	ofn.hInstance         = ghInstance;
	ofn.lpstrFilter       = "Binary card image\0*.*\0Text\0*.txt;\0";
	ofn.lpstrCustomFilter = NULL;
	ofn.nMaxCustFilter    = 0;
	ofn.nFilterIndex      = 1;
	ofn.lpstrFile         = namebuf;
	ofn.nMaxFile          = sizeof(namebuf);
	ofn.lpstrFileTitle    = NULL;
	ofn.lpstrInitialDir   = curdir;							// leaving this NULL does NOT start you in the current directory
	ofn.lpstrTitle        = NULL;							// (on XP, anyway) so let's name it explicitly
	ofn.Flags  		      = OFN_ENABLEHOOK|OFN_EXPLORER;
	ofn.lpfnHook		  = getsave_hook_func;

	if (! GetSaveFileName(&ofn))
		return FALSE;

	as_text = ofn.nFilterIndex == 2;

	if (! write_deck(hDlg, di, namebuf, as_text))
		return FALSE;

	if (! as_text) {										// if they saved as text, don't update name or changed flag
		strcpy(di->curname, namebuf);
		set_window_title(di);
		di->changed = FALSE;
	}

	return TRUE;
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL DeckView_Cmd_File_New (HWND hDlg, DECKINFO *di)
{
	return open_window(NULL, di);
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL DeckView_Cmd_File_Open (HWND hDlg, DECKINFO *di)
{
	OPENFILENAME ofn;
	char namebuf[MAX_PATH], curdir[MAX_PATH];
	BOOL rval;
	int i;

	GetCurrentDirectory(sizeof(curdir), curdir);

	if (! query_close_file(hDlg, di))
		return FALSE;

	strcpy(namebuf, di->curname);
	if ((i = strlen(namebuf)) > 0)
		if (namebuf[i-1] == '\\')
			namebuf[i-1] = '\0';							// dialog doesn't like it if name ends in backslash

	ZeroMemory(&ofn, sizeof(ofn));
	ofn.lStructSize       = sizeof(ofn);
	ofn.hwndOwner         = hDlg;
	ofn.hInstance         = ghInstance;
	ofn.lpstrFilter       = NULL;
	ofn.lpstrCustomFilter = NULL;
	ofn.nMaxCustFilter    = 0;
	ofn.nFilterIndex      = 0;
	ofn.lpstrFile         = namebuf;
	ofn.nMaxFile          = sizeof(namebuf);
	ofn.lpstrFileTitle    = NULL;
	ofn.lpstrInitialDir   = curdir;							// leaving this NULL does NOT start you in the current directory
	ofn.lpstrTitle        = NULL;							// (on XP, anyway) so let's name it explicitly
	ofn.Flags  		      = OFN_PATHMUSTEXIST|OFN_EXPLORER;

	if (! GetOpenFileName(&ofn))
		return FALSE;

	di->changed = di->named = FALSE;
	di->ncards  = 0;
	SendMessage(di->hwDeck, LB_RESETCONTENT, 0, 0);
	set_window_title(di);

	rval = load_deck(hDlg, di, namebuf);
	add_end_of_deck(di);
	return rval;
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

void modifying_deck (DECKINFO *di, BOOL modifying)
{
	if (modifying) 
		SendMessage(di->hwDeck, WM_SETREDRAW, FALSE, 0);	// disable window updates during changes
	else {
		SendMessage(di->hwDeck, WM_SETREDRAW, TRUE, 0);		// reenable window updates
		InvalidateRect(di->hwDeck, NULL, TRUE);				// repaint the window
	}
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL load_deck (HWND hDlg, DECKINFO *di, char *fname)
{
	char errbuf[MAX_PATH+100];
	int i, nread;
	BOOL is_deck = TRUE;
	BOOL more = FALSE;
	FILE *fd;
	extern int errno;
	extern char *_sys_errlist[];
	USHORT *cdata;

	if ((fd = fopen(fname, "rb")) == NULL) {
		sprintf(errbuf, "%s: %s", fname, _sys_errlist[errno]);
		MessageBox(hDlg, errbuf, NULL, MB_OK|MB_ICONEXCLAMATION);
		return FALSE;
	}

	nread = fread(di->cardbuf, 1, 160*MAXCARDS, fd);
	more  = fread(errbuf, 1, 1, fd) > 0;
	fclose(fd);

	is_deck = (nread % 160) == 0;
	if (is_deck) {
		for (i = nread*80, cdata = di->cardbuf; --i >= 0; ) {
			if (*cdata++ & 0x000F) {
//				is_deck = FALSE;
				break;
			}
		}
	}

	if (! is_deck) {
		sprintf(errbuf, "%s is not a binary card image file", fname);
		MessageBox(hDlg, errbuf, NULL, MB_OK | MB_ICONEXCLAMATION);
		set_window_title(di);
		return FALSE;
	}

	di->editable = TRUE;
	if (more) {
		sprintf(errbuf, "Deck is too large, only first %d cards displayed. Editing not allowed.", MAXCARDS);
		MessageBox(hDlg, errbuf, "", MB_ICONEXCLAMATION|MB_OK);
		di->editable = FALSE;
	}

	nread /= 160;

	strcpy(di->curname, fname);
	di->named   = TRUE;
	di->ncards  = nread;
	set_window_title(di);

	di->cardno = 0;
	redisplay_deck(di);
	set_window_title(di);
	return TRUE;
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

void redisplay_deck (DECKINFO *di)
{
	int i;
	char cardtext[90];

	modifying_deck(di, TRUE);											// disable window updates during additions

	SendMessage(di->hwDeck, LB_RESETCONTENT, 0, 0);						// zap the list

	for (i = 0; i < di->ncards; i++) {									// load all of deck into the list box
		convert_card(di->cur_encoding, &di->cardbuf[80*i], cardtext);
		SendMessage(di->hwDeck, LB_ADDSTRING, 0, (LPARAM) cardtext);
	}

	modifying_deck(di, FALSE);											// reenable window updates

	SendMessage(di->hwDeck, LB_SETCARETINDEX, di->cardno, FALSE);

	show_card(di->hDlg, di);
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

void add_end_of_deck (DECKINFO *di)
{
	char *str, cardtext[90];
	int i, j;

	for (i = 0; i < 40; i++)
		cardtext[i] = '\xBB';
	for (; i < 80; i++)
		cardtext[i] = '\xAB';
	cardtext[i] = '\0';

	str = " end of deck ";
	for (i = 40-strlen(str)/2, j = 0; str[j] != '\0'; j++, i++)
		cardtext[i] = str[j];
	
	SendMessage(di->hwDeck, LB_ADDSTRING, 0, (LPARAM) cardtext);
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL DeckView_Cmd_Edit_Cut (HWND hDlg, DECKINFO *di)
{
	if (! copy_selection_to_clipboard(hDlg, di))
		return FALSE;

	return delete_selection(hDlg, di);
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL DeckView_Cmd_Edit_Copy (HWND hDlg, DECKINFO *di)
{
	return copy_selection_to_clipboard(hDlg, di);
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL DeckView_Cmd_Edit_Paste (HWND hDlg, DECKINFO *di)
{
	int selindex[2];

	if (get_selected_items(di, selindex, 2) > 0)
		if (! delete_selection(hDlg, di))
			return FALSE;

	return copy_clipboard_to_deck(hDlg, di);
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL DeckView_Cmd_Edit_Delete (HWND hDlg, DECKINFO *di)
{
	return delete_selection(hDlg, di);
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL DeckView_Cmd_Edit_SelectAll (HWND hDlg, DECKINFO *di)
{
	if (di->ncards > 0)
		SendMessage(di->hwDeck, LB_SELITEMRANGEEX, 0, di->ncards-1);
	else
		SendMessage(di->hwDeck, LB_SELITEMRANGEEX, 1, 0);

	return TRUE;
}

// -----------------------------------------------------------------
// DeckView_Cmd_NextSeq - jump to next card whose sequence
// number is out of sequence relative to current card
// -----------------------------------------------------------------

int get_seq (DECKINFO *di, int i)
{
	USHORT *cp = &di->cardbuf[80*i];
	int col, seq = 0, ch;

	for (col = 79; col >= 72; col--) {
		ch = hollerith_to_ascii(di->cur_encoding, (USHORT) (cp[col] & 0xFFF0));
		if (! isdigit(ch))
			break;
	}

	for (col++; col < 80; col++) {
		ch = hollerith_to_ascii(di->cur_encoding, (USHORT) (cp[col] & 0xFFF0));
		seq = seq*10 + ch-'0';
	}

	return seq;
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL jump_to_line (DECKINFO *di, int i)
{
	if (! BETWEEN(i, 0, di->ncards))
		return FALSE;

	SendMessage(di->hwDeck, LB_SETSEL, (WPARAM) FALSE, (LPARAM)(UINT) -1);	// cancel all selections
	di->cardno = i;												// remember this line number
	SendMessage(di->hwDeck, LB_SETCARETINDEX, i, 0L);			// scroll to and put caret in desired line
	show_card(di->hDlg, di);									// display, as no LBN_SELCHANGE will be sent

	return TRUE;
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL DeckView_Cmd_NextSeq (HWND hDlg, DECKINFO *di)
{
	int selindex[2], seq, i;

	if (get_selected_items(di, selindex, 2) >= 2) {		// doesn't make sense if there are 2 or more lines selected
		MessageBeep(0);
		return FALSE;
	}

	i = di->cardno;
	if ((seq = get_seq(di, i)) <= 0) {					// no sequence # on current card, find one
		while (++i < di->ncards)
			if (get_seq(di, i) > 0)
				break;
	}
	else {
		while (++i < di->ncards)						// look for unexpected seq, or end of deck
			if (get_seq(di, i) != ++seq)
				break;
	}

	jump_to_line(di, i);
	return TRUE;
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

void GoTo_OnCommand (HWND hDlg, int id, HWND hWndCtl, UINT codeNotify)
{
	switch (id) {
		case IDOK:
			EndDialog(hDlg, GetDlgItemInt(hDlg, IDC_LINE, NULL, FALSE));
			break;

		case IDCANCEL:
			EndDialog(hDlg, -1);
			break;
	}
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL GoTo_OnInitDialog (HWND hDlg, HWND hwndFocus, LPARAM lParam)
{
	CenterMe(hDlg);
	return TRUE;
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL CALLBACK GoToDlgProc (HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {
		case WM_INITDIALOG:
			return HANDLE_WM_INITDIALOG(hDlg, wParam, lParam, GoTo_OnInitDialog);

		case WM_COMMAND:
			HANDLE_WM_COMMAND(hDlg, wParam, lParam, GoTo_OnCommand);
			break;

		default:
			return FALSE;
	}

	return TRUE;
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL DeckView_Cmd_Goto (HWND hDlg, DECKINFO *di)
{
	int i;

	if ((i = DialogBox(ghInstance, MAKEINTRESOURCE(IDR_DLG_GOTO), hDlg, GoToDlgProc)) > 0) {
		--i;
		jump_to_line(di, min(i, di->ncards));
	}

	return TRUE;
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

void do_search (DECKINFO *di, BOOL skip)
{
	int i, firsti;
	BOOL found;
	char text[100];

	firsti = i = di->cardno;
	found  = FALSE;

	if (! di->down)			// for up search, always skip original cursor line
		skip = TRUE;

	do {
		if (i != di->ncards && ! skip) {
			convert_card(di->cur_encoding, &di->cardbuf[80*i], text);
			strupr(text);
			if (strstr(text, di->lpstrFindWhat) != NULL) {
				found = TRUE;
				break;
			}
		}

		if (di->down) {
			if (++i >= di->ncards)			// wrap to top
				i = 0;
		}
		else {
			if (--i < 0)
				i = di->ncards-1;
		}

		skip = FALSE;
	} while (i != firsti);					// stop if we circle completely

	di->found = found;

	if (found)
		jump_to_line(di, i);
	else
		MessageBeep(0);
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL DeckView_OnFindText (HWND hDlg, WPARAM wParam, FINDREPLACE *ft)
{
	DECKINFO *di = (DECKINFO *) GetWindowLong(hDlg, DWL_USER);

	if (ft->Flags & FR_DIALOGTERM) {
		hwFind = NULL;
		return TRUE;
	}

	if (strlen(ft->lpstrFindWhat) <= 0)
		return FALSE;

	strupr(strcpy(di->lpstrFindWhat, ft->lpstrFindWhat));
	di->down = ft->Flags & FR_DOWN;

	do_search(di, FALSE);

	return TRUE;
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL DeckView_Cmd_Edit_Find (HWND hDlg, DECKINFO *di)
{
	static FINDREPLACE fr;
	static char txt[MAX_FINDWHAT];

	if (hwFind != NULL) {
		fr.hwndOwner = hDlg;						// set message target
		ShowWindow(hwFind, SW_SHOWNORMAL);
		return TRUE;
	}

	ZeroMemory(&fr, sizeof(fr));

	fr.lStructSize   = sizeof(fr);
	fr.hwndOwner     = hDlg;
	fr.hInstance     = ghInstance;
	fr.Flags         = FR_DOWN | FR_HIDEMATCHCASE | FR_HIDEWHOLEWORD | FR_NOMATCHCASE | FR_NOWHOLEWORD;
	fr.lpstrFindWhat = txt;
	fr.wFindWhatLen  = sizeof(txt);

	strcpy(txt, di->lpstrFindWhat);					// better use previous string for this window

	hwFind = FindText(&fr);
	return TRUE;
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL DeckView_Cmd_Edit_FindNext (HWND hDlg, DECKINFO *di)
{
	if (! di->found)
		return FALSE;

	do_search(di, TRUE);

	return TRUE;
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

void select_encoding (DECKINFO *di, int i)
{
	int j;

	if (i >= 0 && i < nencodings)
		di->cur_encoding = &encoding[i];
	else if (nencodings > 0) {
		di->cur_encoding = &encoding[0];
		i = 0;
	}
	else {
		di->cur_encoding = NULL;
		i = -1;
	}

	for (j = 0; j < nencodings; j++)
		CheckMenuItem(di->hPopup, j, MF_BYPOSITION | ((j == i) ? MF_CHECKED : MF_UNCHECKED));
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

int DeckView_OnRenderFormat (HWND hDlg, UINT cfFormat)
{
	DECKINFO *di = (DECKINFO *) GetWindowLong(hDlg, DWL_USER);

	if (cfFormat == CF_TEXT)
		return render_text(di);

	return 1;						// unknown format
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

void DeckView_OnRenderAllFormats (HWND hDlg)
{
	DECKINFO *di = (DECKINFO *) GetWindowLong(hDlg, DWL_USER);

	render_text(di);
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

void DeckView_OnDestroyClipboard (HWND hDlg)
{
	DECKINFO *di = (DECKINFO *) GetWindowLong(hDlg, DWL_USER);

	di->hClipboardData = NULL;
}

// -----------------------------------------------------------------
// render_text - turn clipboard data into text. Return 0 if successful.
// -----------------------------------------------------------------

int render_text (DECKINFO *di)
{
	LPSTR pText;
	HANDLE hText;
	int i, nb;
	char cardtext[90];
	USHORT *pData;

	if (di->hClipboardData == NULL)
		return 1;

	if ((pData = (USHORT *) GlobalLock(di->hClipboardData)) == NULL)			// failed
		return 1;

	if (di->nClipboardLines != *pData++) {						// first word of clipboard data should be # of cards selected
		GlobalUnlock(di->hClipboardData);
		MessageBeep(0);
		return 1;
	}

	nb = di->nClipboardLines*82 + 10;							// most room the text could take up

	if ((hText = GlobalAlloc(GMEM_MOVEABLE|GMEM_DDESHARE, nb)) == NULL) {
		GlobalUnlock(di->hClipboardData);
		MessageBeep(0);
		return 1;
	}

	if ((pText = GlobalLock(hText)) == NULL) {
		GlobalFree(hText);
		GlobalUnlock(di->hClipboardData);
		MessageBeep(0);
		return 1;
	}

	for (i = 0; i < di->nClipboardLines; i++, pData += 80) {
		convert_card(di->cur_encoding, pData, cardtext);
		nb = strlen(cardtext);
		CopyMemory(pText, cardtext, nb);
		pText += nb;
		*pText++ = '\r';
		*pText++ = '\n';
	}
	*pText = '\0';

	GlobalUnlock(hText);
	SetClipboardData(CF_TEXT, hText);
	return 0;
}


// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL copy_selection_to_clipboard (HWND hDlg, DECKINFO *di)
{
	int i, selindex[MAXCARDS], nsel;
	HANDLE hData;
	USHORT *pData;

	if ((nsel = get_selected_items(di, selindex, MAXCARDS)) <= 0)
		return FALSE;

	if ((hData = GlobalAlloc(GMEM_MOVEABLE|GMEM_DDESHARE, nsel*160+sizeof(USHORT))) == NULL) {
		MessageBox(hDlg, "Unable to allocate memory for clipboard", NULL, MB_ICONEXCLAMATION|MB_OK);
		return FALSE;
	}

	if ((pData = (USHORT *) GlobalLock(hData)) == NULL) {
		MessageBox(hDlg, "Unable to lock memory for clipboard", NULL, MB_ICONEXCLAMATION|MB_OK);
		GlobalFree(hData);
		return FALSE;
	}

	*pData++ = nsel;										// this limits us to 65535 pasted cards!

	for (i = 0; i < nsel; i++, pData += 80)					// copy data for selected cards 
		CopyMemory(pData, di->cardbuf+80*selindex[i], 160);

	GlobalUnlock(hData);

	if (! OpenClipboard(hDlg)) {
		MessageBox(hDlg, "Unable to open clipboard", NULL, MB_ICONEXCLAMATION|MB_OK);
		GlobalFree(hData);
		return FALSE;
	}

	EmptyClipboard();
	SetClipboardData(cfCardFormat, hData);
	SetClipboardData(CF_TEXT, NULL);			// offer to render text for other apps
	CloseClipboard();

	di->hClipboardData  = hData;				// remember the data handle
	di->nClipboardLines = nsel;					// and the number of lines
	return TRUE;
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL copy_clipboard_to_deck (HWND hDlg, DECKINFO *di)
{
	UINT format;
	static UINT formats[2] = {0, CF_TEXT};
	HANDLE hData;
	LPSTR pData;
	USHORT *pDest, *pCardData, cardin[80];
	int i, nadd, jdest, col, xx;
	BOOL overflow = FALSE, added = FALSE;
	char cardtext[90];

	if (! OpenClipboard(hDlg)) {
		MessageBeep(0);
		return FALSE;
	}

	formats[0] = cfCardFormat;
	format = (UINT) GetPriorityClipboardFormat(formats, 2);
	if (format == 0 || format == (UINT) -1) {
		CloseClipboard();
		MessageBeep(0);
		return FALSE;
	}

	if ((hData = GetClipboardData(format)) == NULL) {
		CloseClipboard();
		MessageBeep(0);
		return FALSE;
	}

	if ((pData = GlobalLock(hData)) == NULL) {
		CloseClipboard();
		MessageBeep(0);
		return FALSE;
	}		

	modifying_deck(di, TRUE);

	jdest = di->cardno;
	assert(jdest >= 0);
	assert(jdest <= di->ncards);

	pDest = di->cardbuf + jdest*80;

	if (format == cfCardFormat) {
		pCardData = (USHORT *) pData;

		nadd = *pCardData++;
		if ((di->ncards+nadd) > MAXCARDS) {
			overflow = TRUE;
			nadd = MAXCARDS - di->ncards;
		}

		if (nadd > 0) {
			MoveMemory(pDest+nadd*80, pDest, (di->ncards-jdest)*160);
			CopyMemory(pDest, pCardData, nadd*160);
			di->ncards += nadd;
			added = TRUE;

			for (i = 0; i < nadd; i++, pDest += 80) {
				convert_card(di->cur_encoding, pDest, cardtext);
				SendMessage(di->hwDeck, LB_INSERTSTRING, jdest++, (LPARAM) cardtext);
			}
		}
	}
	else if (format == CF_TEXT) {
		while (*pData) {
			if (di->ncards >= MAXCARDS) {
				overflow = TRUE;
				break;
			}

			FillMemory(cardin, 160, 0);

			for (col = 0; col < 80 && *pData != '\r' && *pData; col++)
				cardin[col] = ascii_to_hollerith(di->cur_encoding, *pData++);

			if (*pData == '\r')
				if (*++pData == '\n')
					pData++;

			xx = (pDest - di->cardbuf)/80;


			MoveMemory(pDest+80, pDest, (di->ncards-jdest)*160);
			CopyMemory(pDest, cardin, 160);

			di->ncards++;
			added = TRUE;

			convert_card(di->cur_encoding, pDest, cardtext);
			SendMessage(di->hwDeck, LB_INSERTSTRING, jdest++, (LPARAM) cardtext);

			pDest += 80;
		}
	}

	di->cardno = jdest;
	SendMessage(di->hwDeck, LB_SETCARETINDEX, jdest, FALSE);
	show_card(hDlg, di);

	modifying_deck(di, FALSE);

	GlobalUnlock(hData);
	CloseClipboard();

	if (added)
		di->changed = TRUE;

	if (overflow)
		MessageBox(hDlg, "Too many cards, could not paste it all in", NULL, MB_ICONEXCLAMATION|MB_OK);
	else if (! added)
		MessageBeep(0);

	return added;
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

int selsort (const void *a, const void *b)
{
	return (* (int *) a > * (int *) b) ? 1 : (* (int *) a < * (int *) b) ? -1 : 0;
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

int get_selected_items (DECKINFO *di, int *selindex, int maxsel)
{
	int nsel;

	nsel = SendMessage(di->hwDeck, LB_GETSELITEMS, maxsel, (LPARAM) selindex);
	if (nsel <= 0)
		return 0;

	qsort(selindex, nsel, sizeof(int), selsort);

	if (selindex[nsel-1] >= di->ncards)			// if last item is the end-of-deck line, remove it
		nsel--;

	return nsel;
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------

BOOL delete_selection (HWND hDlg, DECKINFO *di)
{
	int i, j, selindex[MAXCARDS], nsel;

	if ((nsel = get_selected_items(di, selindex, MAXCARDS)) <= 0)
		return FALSE;

	modifying_deck(di, TRUE);										// disable window updates during deletions

	for (i = nsel; --i >= 0; ) {									// delete bottom up so indexes don't change on us
		j = selindex[i];
		if (j >= di->ncards)										// don't count 'end of deck'
			continue;
		SendMessage(di->hwDeck, LB_DELETESTRING, j, 0);				// delete string from list box
		di->ncards--;												// decrement card count
		CopyMemory(&di->cardbuf[j*80], &di->cardbuf[(j+1)*80], (di->ncards-j)*160);				// slide remaining cards down in the card buffer
	}
																	// select line after last deleted line
	di->cardno = j;
	SendMessage(di->hwDeck, LB_SETCARETINDEX, j, FALSE);
	show_card(hDlg, di);

	modifying_deck(di, FALSE);										// reenable window updates
	SetDlgItemInt(hDlg, IDC_NCARDS, di->ncards, FALSE);				// update displayed card count
	di->changed = TRUE;												// mark deck as changed
	return TRUE;
}

// -----------------------------------------------------------------
// load_reg_bool - load a boolean value from the registry
// -----------------------------------------------------------------

void load_reg_bool (HKEY hKey, LPSTR valname, BOOL *val, BOOL default_val)
{
	DWORD rtype, dval, len;

	if (hKey != INVALID_HANDLE_VALUE) {
		len = sizeof(dval);
		if (RegQueryValueEx(hKey, valname, NULL, &rtype, (LPBYTE) &dval, &len) == ERROR_SUCCESS) {
			if (rtype == REG_DWORD) {
				*val = dval;
				return;
			}
		}
	}

	*val = default_val;
}

// -----------------------------------------------------------------
// load_reg_str - load a string value from the registry
// -----------------------------------------------------------------

void load_reg_str (HKEY hKey, LPSTR valname, LPSTR val, int maxval, LPSTR default_val)
{
	DWORD rtype, len;

	if (hKey != INVALID_HANDLE_VALUE) {
		len = maxval;
		if (RegQueryValueEx(hKey, valname, NULL, &rtype, val, &len) == ERROR_SUCCESS) {
			if (rtype == REG_SZ)
				return;
		}
	}

	strncpy(val, default_val, maxval);
}

// -----------------------------------------------------------------
// save_reg_bool - save a boolean value to the registry
// -----------------------------------------------------------------

void save_reg_bool (HKEY hKey, LPSTR valname, BOOL val)
{
	DWORD dval;

	if (hKey != INVALID_HANDLE_VALUE) {
		dval = val;
		RegSetValueEx(hKey, valname, 0, REG_DWORD, (CONST BYTE *) &dval, sizeof(DWORD));
	}
}

// -----------------------------------------------------------------
// save_reg_str - save a string value to the registry
// -----------------------------------------------------------------

void save_reg_str (HKEY hKey, LPSTR valname, LPSTR val)
{
	if (hKey != INVALID_HANDLE_VALUE) {
		RegSetValueEx(hKey, valname, 0, REG_SZ, val, strlen(val)+1);
	}
}

// -----------------------------------------------------------------
// locate_program - find a program specified by name and get its full path
// -----------------------------------------------------------------

BOOL locate_program (char *name, char *fullpath, int s_fullpath)
{
	char *basename;

	if (strchr(name, ':') != NULL || strchr(name, '\\') != NULL) {	// user specified an explicit path
		strcpy(fullpath, name);										// use that
	}
	else if (SearchPath(NULL, name, NULL, s_fullpath, fullpath, &basename) == 0) {
		*fullpath = '\0';											// ... not found. zap string and return
		return FALSE;
	}

	// find the program

	if (GetFileAttributes(fullpath) != 0xFFFFFFFF) 					// but be sure file exists
		return TRUE;												// ... it does

	*fullpath = '\0';												// ... it doesn't; zap the path string
	return FALSE;
}

// -----------------------------------------------------------------
// get_defaults - load default values from the registry
// -----------------------------------------------------------------

void get_defaults (char *dir, int s_dir)
{
	HKEY hkReg;

	if (RegOpenKey(HKEY_CURRENT_USER, REGISTRY_KEY, &hkReg) != ERROR_SUCCESS)
		hkReg = INVALID_HANDLE_VALUE;

	load_reg_str(hkReg, "directory", dir, s_dir, "");
	load_reg_str(hkReg, "encoding_file", encoding_file, sizeof(encoding_file), DEFAULT_ENCODING_FILE);
	load_reg_str(hkReg, "encoding",  encoding_name, sizeof(encoding_name), DEFAULT_ENCODING_NAME);

	if (hkReg != INVALID_HANDLE_VALUE)
		RegCloseKey(hkReg);

	locate_program(encoding_file, encoding_file_path, sizeof(encoding_file_path));
}

// -----------------------------------------------------------------
// save_defaults - store all GUI settings to registry for next time
// -----------------------------------------------------------------

void save_defaults (void)
{
	HKEY hkReg;
	char dir[MAX_PATH];
	DWORD dispos;

	GetCurrentDirectory(sizeof(dir), dir);

	if (RegCreateKeyEx(HKEY_CURRENT_USER, REGISTRY_KEY, 0, "",
			REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &hkReg, &dispos) != ERROR_SUCCESS)
		hkReg = INVALID_HANDLE_VALUE;

	save_reg_str(hkReg, "directory", dir);
	save_reg_str(hkReg, "encoding",  encoding_name);		// save last-selected encoding

	if (hkReg != INVALID_HANDLE_VALUE)
		RegCloseKey(hkReg);
}

// -----------------------------------------------------------------
// show_card - update card image with current data
// -----------------------------------------------------------------

void show_card (HWND hDlg, DECKINFO *di)
{
	if (BETWEEN(di->cardno, 0, di->ncards-1)) {
		SetDlgItemInt(hDlg, IDC_CARDNO, di->cardno+1, FALSE);
		SetDlgItemText(hDlg, IDC_OF, "of");
	}
	else {
		SetDlgItemText(hDlg, IDC_CARDNO, "");
		SetDlgItemText(hDlg, IDC_OF, "");
	}
	SetDlgItemInt(hDlg, IDC_NCARDS, di->ncards, FALSE);

	InvalidateRect(di->hwCard, NULL, FALSE);					// force repaint of the card image (no need to erase, paint_card floods the rectangle)
}

// -----------------------------------------------------------------
// MyRectProc - superclass window procedure for dialog box "groupbox".
// This is used to intercept WM_PAINT message to the groupbox representing the card image
// -----------------------------------------------------------------

LRESULT CALLBACK MyRectProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	LRESULT rval;

	rval = (*oldRectProc)(hWnd, uMsg, wParam, lParam);
	
	if (uMsg == WM_PAINT)
		paint_card((DECKINFO *) GetWindowLong(GetParent(hWnd), DWL_USER));

	return rval;
}

// -----------------------------------------------------------------
// paint_card - paint the current cardbuf data in the IDC_CARD box; called
// after WM_PAINT is handled by the original control's window procedure.
// (This isn't a perfect situation, as it doesn't get to use the BeginPaint
// display context; it gets its own and so paints the entire region. But that
// isn't a terrible inefficiency as most paints will be the full region anyway).
// If verify is on, we paint red to show differences between the two decks:
// red punches where there was none in the file, and red outlines where
// there was a punch in the file but not in the current card
// -----------------------------------------------------------------

void paint_card (DECKINFO *di)
{
	HDC hDC;
	RECT rCard, rBar, rHole;
	int hCard, wCard, xCard, yCard, row, col, cdata, x, y[12], whole, hhole, dx, bit;
	USHORT *card = &di->cardbuf[80*di->cardno];

	GetWindowRect(di->hwCard, &rCard);						// get screen coordinates for card and text boxes
	rBar.top    = rCard.top;
	rBar.bottom = rCard.top + 25;

	hCard = rCard.bottom-rCard.top-(rBar.bottom-rBar.top)-1;// height of drawable space in card
	wCard = rCard.right-rCard.left-1;

	hDC = GetDC(di->hwCard);								// allocate a display context

	whole = (int) (wCard/80.*0.8) - 1;						// width and height of punch rectangles
	hhole = (int) (hCard/12.*0.7);

	xCard = 3;												// position of punch for column 1 row 12
	yCard = rBar.bottom - rCard.top + 2;

	rCard.left   = 1;
	rCard.right  = wCard - 1;
	rCard.bottom = rCard.bottom - rCard.top - 2;
	rCard.top    = 8;

	if (bManilla != NULL && di->cardno < di->ncards)		// paint the card region a nice manilla color
		FillRect(hDC, &rCard, bManilla);
	else
		FillRect(hDC, &rCard, GetStockObject(GRAY_BRUSH));

	if (BETWEEN(di->cardno, 0, di->ncards-1)) {
		for (row = 0; row < 12; row++)						// precompute row y coordinates
			y[row] = yCard + (int) (row*hCard/12);

		for (col = 0; col < 80; col++) {					// now, check every bit for a punch
			x = xCard + (int) (wCard/80.*col);
			cdata = *card++;
			bit = 0x8000;
			for (row = 0; row < 12; row++, bit >>= 1) {
				if (cdata & bit) {
					rHole.left   = x;
					rHole.right  = x + whole + 1;
					rHole.top    = y[row];
					rHole.bottom = y[row] + hhole + 1;
					FillRect(hDC, &rHole, bBlack);
				}
			}
			x += dx;
		}
	}

	ReleaseDC(di->hwCard, hDC);								// release the DC
}

// -----------------------------------------------------------------
// convert_card - convert binary data in cardbuf to text string in cardtext
// -----------------------------------------------------------------

void convert_card (ENCODING *encoding, USHORT *card, char *cardtext)
{
	int col, lastcol = -1;

	for (col = 0; col < 80; col++) {						// get ascii equivalent for each character
		cardtext[col] = hollerith_to_ascii(encoding, (USHORT) (*card++ & 0xFFF0));
		if (cardtext[col] > ' ')							// remember index of last nonblank character
			lastcol = col;
	}

	cardtext[lastcol+1] = '\0';								// terminate string after last nonblank character
}

// -----------------------------------------------------------------
// skipbl - skip leading whitespace in a string
// -----------------------------------------------------------------

char *skipbl (char *str)
{
	while (*str && *str <= ' ')
		str++;

	return str;
}

// -----------------------------------------------------------------
// astring - allocate space and return a copy of a string
// -----------------------------------------------------------------

char *astring (char *str)
{
	char *s;

	if ((s = malloc(strlen(str)+1)) == NULL) {
		bail("Insufficient memory");
		return "";
	}

	strcpy(s, str);
	return s;
}

// -----------------------------------------------------------------
// CenterMe - center calling window in its parent.  Call this from WM_INITDIALOG
// message for Dialog Boxes which we want centered on the screen.
// -----------------------------------------------------------------

void CenterMe (HWND hWnd)
{
	POINT pt;
	RECT  swp, rParent, rScreen;
	int   iwidth, iheight;
	HWND  hwParent, hwDesktop = GetDesktopWindow();

	if ((hwParent = GetParent(hWnd)) == NULL)
    	hwParent = hwDesktop;
	
	GetClientRect(hwParent, &rParent);						// get dimensions of parent window client area
	pt.x = rParent.right/ 2;								// find the center point in screen coordinates
	pt.y = rParent.bottom/ 2;
	ClientToScreen(hwParent, &pt);

	GetWindowRect(hWnd, &swp);								// get bounding rectangle of dialog
	iwidth  = swp.right - swp.left;							// calculate the height and width
	iheight = swp.bottom - swp.top;

	pt.x = pt.x - (iwidth / 2);								// put dialog center on center of parent client
	pt.y = pt.y - (iheight / 2);

	GetWindowRect(hwDesktop, &rScreen);						// get screen dimensions

	GetWindowRect(hwParent, &rParent);	

	if ((pt.x+iwidth) > rParent.right)						// if that would put part of dialog off
		pt.x = rParent.right-iwidth;						// the parent, move it back
	if (pt.x < rParent.left)
		pt.x = 0;

	if ((pt.y+iheight) > rParent.bottom)
		pt.y = rParent.bottom-iheight;
	if (pt.y < rParent.top)
		pt.y = 0;
															// move it to chosen position
	MoveWindow(hWnd, pt.x, pt.y, iwidth, iheight, FALSE);
}

// -----------------------------------------------------------------
// bail - report fatal error and close the application
// -----------------------------------------------------------------

void bail (char *msg)
{
	MessageBox(IsWindow(hwTop) ? hwTop : GetDesktopWindow(), msg, NULL, MB_ICONASTERISK);		// display message
	PostQuitMessage(1);
}

void _my_assert(char *expr, char *filename, unsigned lineno)
{
	char asmsg[250];

	sprintf(asmsg, "Assertion %s failed, %s:%d", expr, filename, lineno);
	bail(asmsg);
}

